import pygame, sys, os, random
from pygame.locals import *

#Game Constants
WINDOWHEIGHT = 640
WINDOWWIDTH = 640

ACCELERATION = 0.2
MAX_SPEED = 2
GRAVITY = 0.1
DEACCELERATION = 0.4
JUMP = 4.5
SIZE = 32
NEWFRAME = 20

FPS = 100

#Game variables
def variables():
    global player_pos, player_v, up, down, right, left, jump, facing, frame
    global frame_tick, gamewin, levelbegin, platformlist, hazardlist
    global exitlist, levelmessagesurface, levelmessagerect, messagelist
    global backgrounsurface, backgroundrect, nextlevel, messageindex
    global player_behaviour, player_antimatter, player_normal
    global playersurface, playerrect, playerlist, updatescreen

    up = False
    down = False
    right = False
    left = False
    jump = False
    facing = 'left'
    frame = {}
    frame_tick = {}
    gamewin = False
    levelbegin = True
    platformlist = []
    hazardlist = []
    exitlist = []
    imagereference = {}
    levelmessagesurface = None
    levelmessagerect = Rect(0,0,0,0)
    messagelist = []
    messageindex = 0
    backgroundsurface = None
    backgroundrect = Rect(0,0,0,0)
    nextlevel = None
    updatescreen = True

    player_v = {}
    player_pos = {'player1':(0,0)}
    playersurface = {}
    playerrect = {}
    playerlist = []

#function to load images files return image file and rect
#Arg1 name, Arg2 transparency value
def load_image(name, colourkey=None):
    fullname = os.path.join('Images',name)
    try:
        image = pygame.image.load(fullname)
    except pygame.error, message:
        print 'Cannot load image:', name
        raise SystemExit, message
    if colourkey is not None:
        if colourkey is -1:
            colourkey = image.get_at((0,0))
        image.set_colorkey(colourkey, RLEACCEL)
        #print name, image.get_at((0,0)), colourkey
    return image, image.get_rect()

#loads files from the Levels folder
def load_levelfile(name):
    levelfile = None
    fullname = os.path.join('Levels',name)
    try:
        levelfile = open(fullname,'r')
    except:
        print 'Cannot load file:',name
    return levelfile

#function to load sound files returns sound file
def load_sound(name):
    class NoneSound:
        def play(self): pass
    if not pygame.mixer:
        return NoneSound()
    fullname = os.path.join('Sounds',name)
    try:
        sound = pygame.mixer.Sound(fullname)
    except pygame.error, message:
        print 'Cannot load sound:', name
        raise SystemExit, messgae
    return sound

#adds vectors
def add_vector(vector1,vector2):
    return (vector1[0] + vector2[0] , vector1[1] + vector2[1])

#calculates position based upon angle and length
def angle_length(angle, length):
    radian_angle = math.radians(angle)
    return (-(length * math.sin (radian_angle)),
            length * math.cos (radian_angle))

#Exits window
def terminate():
    pygame.quit()
    sys.exit()

#updates player on screen
def surface_update(image,rect):
    windowSurface.blit(image,rect)

#changes the position and/or size of a rect
def move_rect(pos,size):
    return Rect(pos[0],pos[1],size[0],size[1])

#Checks to see if player exits window (escape or close button)
def check_quit(event):
    if event.type == QUIT:
        terminate()
    if event.type == KEYDOWN:
        if event.key == K_ESCAPE:
            terminate()

#saves dictionary of images given the file name
def load_levelimages(filename):
    global imagereference
    imagereference = {}
    levelfile = load_levelfile(filename)
    for fileline in levelfile.readlines():
        name, filename = fileline.rstrip().split(':')
        imagereference[name] = filename

def load_animations(filename):
    global animations, imagereference
    animations = {}
    surface = None
    animationfile = load_levelfile(filename)
    for fileline in animationfile.readlines():
        name, files = fileline.rstrip().split(':')
        files = files.split(',')
        animations[name] = []
        for image in files:
            surface, forget = load_image(imagereference[image],(109, 2, 217))
            animations[name].append(surface)

#updates all surfaces on-screen
def load_screen():
    windowSurface.blit(backgroundsurface, backgroundrect)
    for platform in platformlist:
        surface_update(platform[0],platform[1])
    for hazard in hazardlist:
        surface_update(hazard[0],hazard[1])
    for exitblock in exitlist:
        surface_update(exitblock[0],exitblock[1])
    for player in playerlist:
        surface_update(playersurface[player],playerrect[player])

#loads platform, spawn, hazards and other game images from a file
def load_level(filename):
    global platformlist, imagereference, hazardlist, exitlist, nextlevel, levelbegin
    global levelmessagesurface, levelmessagerect, backgroundsurface, backgroundrect
    global messagelist, messageindex, playersurface, playerrect, player_pos
    global playerlist, player_v, frame_tick, frame, facing, currentlevel
    variables()
    levelfile = load_levelfile(filename)
    filelist = levelfile.readlines()
    i = 0
    currentlevel = filename
    while i < len(filelist)-1:
        fileline = filelist[i].rstrip()
        while fileline.endswith(';'):
            fileline = fileline.strip(';')
            i += 1
            fileline += ',' + filelist[i].rstrip()
        blocktype, reference, coordinates = fileline.split(':')
        imagefile, width, height = reference.split(',' )
        coordinatelist = coordinates.replace('(','').replace(')','').split(',')
        blockimage, blockrect = load_image(imagereference[imagefile],
                                           (109, 2, 217))
        i += 1
        if blocktype.startswith('player'):
            for coordinate in coordinatelist:
                x,y = coordinate.split('.')
                player_pos[blocktype] = (int(x),int(y))
                playersurface[blocktype] = blockimage
                playerrect[blocktype] = Rect(int(x),int(y),int(width),int(height))
                playerlist.append(blocktype)
                player_v[blocktype] = (0,0)
                frame_tick[blocktype] = NEWFRAME-1
                frame[blocktype] = 0
        if blocktype == 'levelmessage' or blocktype == 'background':
            for coordinate in coordinatelist:
                x,y = coordinate.split('.')
                if blocktype == 'background':
                    backgroundsurface = blockimage
                    backgroundrect = Rect(int(x),int(y),int(width),int(height))
                if blocktype == 'levelmessage':
                    if len(messagelist) == 0:
                        levelmessagesurface = blockimage
                        levelmessagerect = Rect(int(x),int(y),int(width),
                                                int(height))
                    messagelist.append([blockimage, Rect(int(x),int(y),
                                                         int(width),int(height))])
                     
        if blocktype == 'platform' or blocktype == 'hazard' or\
           blocktype == 'exit':
            for coordinate in coordinatelist:
                xwidth = 1
                yheight = 1
                secx,secy = coordinate.split('.')
                if 'x' in secx: x,xwidth = secx.split('x')
                else: x = secx
                if 'x' in secy: y,yheight = secy.split('x')
                else: y = secy
                for w in range(int(xwidth)):
                    for h in range(int(yheight)):
                        if blocktype == 'platform':
                            platformlist.append([blockimage,
                                                 Rect(int(x)+int(width)*w,
                                                      int(y)+int(height)*h,
                                                      int(width),int(height))])
                        if blocktype == 'hazard':
                            hazardlist.append([blockimage,
                                               Rect(int(x)+int(width)*w,
                                                    int(y)+int(height)*h,
                                                    int(width),int(height))])
                        if blocktype == 'exit':
                            exitlist.append([blockimage,
                                            Rect(int(x)+int(width)*w,
                                                 int(y)+int(height)*h,
                                                 int(width),int(height))])
        #if blocktype == 'background':
            #background = load_image(imagereference[imagefile])[0]
        nextlevel = filelist[-1].rstrip()

def load_behaviours(filename):
    global player_behaviour, player_antimatter, player_normal
    levelfile = load_levelfile(filename)
    player_behaviour = {}
    player_normal = []
    player_antimatter = []
    for fileline in levelfile.readlines():
        name, behaviour, version = fileline.rstrip().split(':')
        player_behaviour[name] = behaviour
        if version == 'normal':
            player_normal.append(name)
        if version == 'antimatter':
            player_antimatter.append(name)
#Checks if player will move out of boundaries and returns True if so
def check_boundary(pos,vector=(0,0)):
    global SIZE, WINDOWHEIGHT, WINDOWWIDTH
    if pos[0]+vector[0] < 0:
        return True, False
    if pos[0]+SIZE+vector[0] > WINDOWWIDTH:
        return True, False
    if pos[1]+vector[1] < 0:
        return False, True
    if pos[1]+SIZE+vector[1] > WINDOWHEIGHT:
        return False, True
    return False, False

#Returns true if rects overlap
def check_collision(rect1, rect2):
    return rect2.collidepoint((rect1.left,rect1.bottom-5))|rect2.collidepoint((rect1.topleft)),\
           rect2.collidepoint((rect1.right,rect1.bottom-5))|rect2.collidepoint((rect1.topright)),\
           rect2.collidepoint(rect1.midbottom),rect2.collidepoint(rect1.midtop)

#moves player based on input
def move_player(player):
    global player_pos, player_v, platformlist, hazardlist, exitlist, frame_tick
    global gamewin, nextlevel, levelbegin, playerrect, player_behaviour
    if player_behaviour[player] == 'nomove':
        return
    
    vx = player_v[player][0]
    vy = player_v[player][1]
    
    if player_behaviour[player] == 'floating' or\
       player_behaviour[player] == 'inversefloating':

        pup = False
        pdown = False
        if player_behaviour[player] == 'inversefloating':
            if down: pup = True
            if up: pdown = True
        else:
            if up: pup = True
            if down: pdown = True
            
        if pup: vy += -ACCELERATION
        elif pdown: vy += ACCELERATION
        elif vy < 0 and vy+DEACCELERATION < 0: vy += DEACCELERATION
        elif vy > 0 and vy-DEACCELERATION > 0: vy += -DEACCELERATION
        else: vy = 0
        if vy > MAX_SPEED: vy = MAX_SPEED
        if vy < -MAX_SPEED: vy = -MAX_SPEED

        #move up or down
        velocity = (vx,vy)
        pos = add_vector(player_pos[player],velocity)
        playerrect[player] = Rect(pos[0],pos[1],SIZE,SIZE)

        #check boundary
        if check_boundary(pos)[0]: vx = 0
        if check_boundary(pos)[1]: vy = 0

        #check for collision of normal and antimatter
        if player in player_normal:
            for antimatter in player_antimatter:
                if check_collision(playerrect[player], playerrect[antimatter])[0] or\
                   check_collision(playerrect[player], playerrect[antimatter])[1] or\
                   check_collision(playerrect[player], playerrect[antimatter])[2] or\
                   check_collision(playerrect[player], playerrect[antimatter])[3]:
                    del playerlist[playerlist.index(player)]
                    del playerlist[playerlist.index(antimatter)]
                    del playerrect[player]
                    del playerrect[antimatter]
                    del player_normal[player_normal.index(player)]
                    del player_antimatter[player_antimatter.index(antimatter)]
                    if len(playerlist) == 0:
                        load_behaviours(currentlevel.strip('.txt')+'behaviour.txt')
                        load_level(currentlevel)
                        for player in playerlist:
                            playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
                        load_screen()
                        surface_update(levelmessagesurface, levelmessagerect)
                        pygame.display.update()
                        levelbegin = True
                    return

        #Check for collision with hazard
        for hazard in hazardlist:
            if check_collision(playerrect[player], hazard[1])[0] or\
               check_collision(playerrect[player], hazard[1])[1] or\
               check_collision(playerrect[player], hazard[1])[2] or\
               check_collision(playerrect[player], hazard[1])[3]:
                del playerlist[playerlist.index(player)]
                del playerrect[player]
                if player in player_normal:del player_normal[player_normal.index(player)]
                else: del player_antimatter[player_antimatter.index(player)]
                if len(playerlist) == 0:
                    load_behaviours(currentlevel.strip('.txt')+'behaviour.txt')
                    load_level(currentlevel)
                    for player in playerlist:
                        playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
                    load_screen()
                    surface_update(levelmessagesurface, levelmessagerect)
                    pygame.display.update()
                    levelbegin = True
                return

        #Check for exit
        for exitblock in exitlist:
            if check_collision(playerrect[player], exitblock[1])[0] or\
               check_collision(playerrect[player], exitblock[1])[1] or\
               check_collision(playerrect[player], exitblock[1])[2] or\
               check_collision(playerrect[player], exitblock[1])[3]:
                if player_behaviour[player] == 'finish':
                    gameoversurface, gameoverrect = load_image(imagereference['win'])
                    surface_update(gameoversurface,gameoverrect)
                    gamewin = True
                    return
                load_behaviours(nextlevel.strip('.txt')+'behaviour.txt')
                load_level(nextlevel)
                for player in playerlist:
                    playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
                load_screen()
                surface_update(levelmessagesurface, levelmessagerect)
                pygame.display.update()
                levelbegin = True
                return
 
        if player in player_normal:
            sametype = player_normal
        if player in player_antimatter:
            sametype = player_antimatter
                    
        for platform in sametype:
            if platform == player: pass
            elif check_collision(playerrect[player], playerrect[platform])[3]:
                player_pos[platform] = add_vector(player_pos[platform],(0,vy))
            elif check_collision(playerrect[player], playerrect[platform])[2]:
                vy = 0
        #check for collision with platform
        for platform in platformlist:
            if len(platform) == 2:
                platform = platform[1]
            if check_collision(playerrect[player], platform)[0]\
               and vx < 0: vx = 0
            if check_collision(playerrect[player], platform)[1]\
               and vx > 0: vx = 0
            if check_collision(playerrect[player], platform)[3]:
                vy = 0
            if check_collision(playerrect[player], platform)[2]:
                vy = 0

        #add vectors
        player_v[player] = (vx,vy)
        player_pos[player] = add_vector(player_pos[player],player_v[player])
        return
        
        
    pright = False
    pleft = False
    if player_behaviour[player] == 'inverse':
        if left: pright = True
        if right: pleft = True
    else:
        if left: pleft = True
        if right: pright = True
        
    #add gravity
    vy += GRAVITY

    #left right up down control
    if pright: vx += ACCELERATION
    if pleft: vx += -ACCELERATION
    
    #reduce speed to max
    if vx > MAX_SPEED: vx = MAX_SPEED
    if vx < -MAX_SPEED: vx = -MAX_SPEED
    
    velocity = (vx,vy)
    pos = add_vector(player_pos[player],velocity)
    playerrect[player] = Rect(pos[0],pos[1],SIZE,SIZE)

    #check if exits screen
    if check_boundary(pos)[0]:
        load_behaviours(currentlevel.strip('.txt')+'behaviour.txt')
        load_level(currentlevel)
        for player in playerlist:
             playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
        load_screen()
        surface_update(levelmessagesurface, levelmessagerect)
        pygame.display.update()
        levelbegin = True
        return
    if check_boundary(pos)[1]:
        load_behaviours(currentlevel.strip('.txt')+'behaviour.txt')
        load_level(currentlevel)
        for player in playerlist:
            playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
        load_screen()
        surface_update(levelmessagesurface, levelmessagerect)
        pygame.display.update()
        levelbegin = True
        return
        
    #check for collision of normal with antimatter
    if player in player_normal:
        for antimatter in player_antimatter:
            if check_collision(playerrect[player], playerrect[antimatter])[0] or\
               check_collision(playerrect[player], playerrect[antimatter])[1] or\
               check_collision(playerrect[player], playerrect[antimatter])[2] or\
               check_collision(playerrect[player], playerrect[antimatter])[3]:
                del playerlist[playerlist.index(player)]
                del playerlist[playerlist.index(antimatter)]
                del playerrect[player]
                del playerrect[antimatter]
                del player_normal[player_normal.index(player)]
                del player_antimatter[player_antimatter.index(antimatter)]
                if len(playerlist) == 0:
                    load_level(currentlevel)
                    load_behaviours(currentlevel.strip('.txt')+'behaviour.txt')
                    for player in playerlist:
                        playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
                    load_screen()
                    surface_update(levelmessagesurface, levelmessagerect)
                    pygame.display.update()
                    levelbegin = True
                return

    #check for collision with hazard
    for hazard in hazardlist:
        if check_collision(playerrect[player], hazard[1])[0] or\
           check_collision(playerrect[player], hazard[1])[1] or\
           check_collision(playerrect[player], hazard[1])[2] or\
           check_collision(playerrect[player], hazard[1])[3]:
            del playerlist[playerlist.index(player)]
            del playerrect[player]
            if player in player_normal:
                del player_normal[player_normal.index(player)]
            else: del player_antimatter[player_antimatter.index(player)]
            if len(playerlist) == 0:
                load_level(currentlevel)
                load_behaviours(currentlevel.strip('.txt')+'behaviour.txt')
                for player in playerlist:
                    playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
                load_screen()
                surface_update(levelmessagesurface, levelmessagerect)
                pygame.display.update()
                levelbegin = True
            return

    #check for collision with exit
    for exitblock in exitlist:
        if check_collision(playerrect[player], exitblock[1])[0] or\
           check_collision(playerrect[player], exitblock[1])[1] or\
           check_collision(playerrect[player], exitblock[1])[2] or\
           check_collision(playerrect[player], exitblock[1])[3]:
            if player_behaviour[player] == 'finish':
                gameoversurface, gameoverrect = load_image(imagereference['win'])
                surface_update(gameoversurface,gameoverrect)
                gamewin = True
                return
            load_behaviours(nextlevel.strip('.txt')+'behaviour.txt')
            load_level(nextlevel)
            for player in playerlist:
                playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
            load_screen()
            surface_update(levelmessagesurface, levelmessagerect)
            pygame.display.update()
            levelbegin = True
            return
        
    sametype = []
    if player in player_normal:
        for platform in player_normal:
            if platform != player:
                sametype.append(playerrect[platform])
    if player in player_antimatter:
        for platform in player_antimatter:
            if platform != player:
                sametype.append(playerrect[platform])
            
    #check for collision with platform
    for platform in platformlist + sametype:
        if len(platform) == 2:
            platform = platform[1]
        if check_collision(playerrect[player], platform)[0]\
           and vx < 0: vx = 0
        if check_collision(playerrect[player], platform)[1]\
           and vx > 0: vx = 0
        if check_collision(playerrect[player], platform)[3]:
            vx = 0
            vy = GRAVITY
        if check_collision(playerrect[player], platform)[2]:
            vy = 0
            if pright != True and vx > 0:
                if vx-DEACCELERATION > 0: vx += -DEACCELERATION
                else: vx = 0
            if pleft != True and vx < 0:
                if vx+DEACCELERATION < 0: vx += DEACCELERATION
                else: vx = 0
            if jump: vy = -JUMP
            if left and frame_tick[player] == NEWFRAME or right and \
               frame_tick[player] == NEWFRAME:
                animate_player(player+'walk', player)
                frame_tick[player] = 0
            if left or right:
                frame_tick[player] += 1
            if left != True and right != True:
                frame_tick[player] = NEWFRAME-1
        


    #add vectors
    player_v[player] = (vx,vy)
    player_pos[player] = add_vector(player_pos[player],player_v[player])

#increase frames for animation
def animate_player(animation, player):
    global frame, animations, playersurface, right
    frame[player] += 1
    if frame[player] >= len(animations[animation]):
        frame[player] = 0
    playersurface[player] = animations[animation][frame[player]]
    if facing == 'right' and right:
        playersurface[player] = pygame.transform.flip(playersurface[player],1,0)

#Check for input and moves player
def check_move(event):
    global up, down, left, right, jump, playersurface, facing
    if event.type == KEYUP:
        if event.key == K_UP: up = False
        if event.key == K_DOWN: down = False
        if event.key == K_RIGHT: right = False
        if event.key == K_LEFT: left = False
        if event.key == K_SPACE: jump = False
    if event.type == KEYDOWN:
        if event.key == K_r:
            load_level(currentlevel)
            load_behaviours(currentlevel.strip('.txt')+'behaviour.txt')
            for player in playerlist:
                playerrect[player] = move_rect(player_pos[player],(SIZE,SIZE))
            load_screen()
            surface_update(levelmessagesurface, levelmessagerect)
            pygame.display.update()
            levelbegin = True
        if event.key == K_UP: up = True
        if event.key == K_DOWN: down = True
        if event.key == K_RIGHT:
            right = True
            for player in playerlist:
                if facing == 'left':
                    playersurface[player] = pygame.transform.flip(playersurface[player],1,0)
                    facing = 'right'
        if event.key == K_LEFT:
            left = True
            for player in playerlist:
                if facing == 'right':
                    playersurface['player1'] = pygame.transform.flip(playersurface[player],1,0)
                    facing = 'left'
            
        if event.key == K_SPACE: jump = True
    
#Other functions

#set up pygame and window
pygame.init()
mainClock = pygame.time.Clock()
windowSurface = pygame.display.set_mode((WINDOWWIDTH,WINDOWHEIGHT))
pygame.display.set_caption('Super Dimension')

#load images
load_levelimages('images.txt')
load_animations('animation.txt')
load_level('level1.txt')
load_behaviours('level1behaviour.txt')

#Event loop, manages controls and updates images etc.
load_screen()
surface_update(levelmessagesurface, levelmessagerect)
pygame.display.update()

while True:
    for event in pygame.event.get():
        check_quit(event)
        if levelbegin and event.type == KEYDOWN and event.key == K_RETURN:
            if messageindex == len(messagelist)-1:
                levelbegin = False
            else:
                messageindex += 1
                levelmessagesurface, levelmessagerect = messagelist[messageindex]
                load_screen()
                surface_update(levelmessagesurface, levelmessagerect)
        elif levelbegin != True:
            check_move(event)
        if gamewin and event.key == K_RETURN:
            terminate()
    if gamewin:
        gameoversurface, gameoverrect = load_image(imagereference['win'])
        surface_update(gameoversurface,gameoverrect)
    elif levelbegin:
        pass
    else:
        for player in playerlist:
            if levelbegin != True and player in playerlist:
                move_player(player)
        if levelbegin: pass
        else:
            load_screen()
    pygame.display.update()
    mainClock.tick(FPS)
